package ru.dbotthepony.kommons.arrays

import java.nio.ByteBuffer
import java.util.BitSet

abstract class Boolean2DArray : Array2D() {
	abstract operator fun get(column: Int, row: Int): Boolean
	abstract operator fun set(column: Int, row: Int, value: Boolean)

	fun storeColumnRow(target: ByteBuffer) = matrixStoreColumnRow(this, target, Boolean2DArray::get) { a, b -> a.put(if (b) 1 else 0) }
	fun storeRowColumn(target: ByteBuffer) = matrixStoreRowColumn(this, target, Boolean2DArray::get) { a, b -> a.put(if (b) 1 else 0) }
	fun store(target: Boolean2DArray) { target.load(this) }

	fun loadColumnRow(target: ByteBuffer) = matrixLoadColumnRow(this, target, Boolean2DArray::set) { it.get() > 0 }
	fun loadRowColumn(target: ByteBuffer) = matrixLoadRowColumn(this, target, Boolean2DArray::set) { it.get() > 0 }
	fun load(from: Boolean2DArray) { matrixLoad(from, this, Boolean2DArray::get, Boolean2DArray::set) }
	fun storeTransposed(from: Boolean2DArray) { matrixLoadTransposed(from, this, Boolean2DArray::get, Boolean2DArray::set) }

	open fun contentEquals(other: Boolean2DArray): Boolean = matrixContentEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }

	open fun fill(value: Boolean) {
		for (x in 0 until rows) {
			for (y in 0 until columns) {
				this[x, y] = value
			}
		}
	}

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this, Boolean2DArray::get)
	}

	open val transposed: Boolean2DArray get() {
		return allocate(rows, columns).also { it.load(this) }
	}

	private class Impl(
		override val columns: Int,
		override val rows: Int,
		private val mem: BitSet,
	) : Boolean2DArray() {
		constructor(
			columns: Int,
			rows: Int,
		) : this(columns, rows, BitSet(columns * rows))

		constructor(other: Impl) : this(other.columns, other.rows, other.mem.clone() as BitSet)

		override fun get(column: Int, row: Int): Boolean {
			if (column !in 0 until columns || row !in 0 until rows)
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")

			return mem[column + row * columns]
		}

		override fun set(column: Int, row: Int, value: Boolean) {
			if (column !in 0 until columns || row !in 0 until rows)
				throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")

			mem[column + row * columns] = value
		}

		override fun fill(value: Boolean) {
			if (value)
				mem.set(0, columns * rows)
			else
				mem.clear(0, columns * rows)
		}

		override fun equals(other: Any?): Boolean {
			if (other is Impl) return mem == other.mem
			return super.equals(other)
		}

		override fun hashCode(): Int {
			return mem.hashCode().rotateLeft(13) + columns * 31 + rows * 31 * 31
		}

		override fun toString(): String {
			return "Boolean2DArray" + super.toString()
		}
	}

	private class View(val parent: Boolean2DArray) : Boolean2DArray() {
		override val rows: Int
			get() = parent.rows
		override val columns: Int
			get() = parent.columns

		override fun get(column: Int, row: Int): Boolean {
			return parent[column, row]
		}

		override fun set(column: Int, row: Int, value: Boolean) {
			throw UnsupportedOperationException("Read only view")
		}

		override fun equals(other: Any?): Boolean {
			return other is View && parent == other.parent
		}

		override fun hashCode(): Int {
			return parent.hashCode()
		}

		override fun toString(): String {
			return "View = $parent"
		}
	}

	companion object {
		@JvmStatic
		fun allocate(columns: Int, rows: Int): Boolean2DArray {
			return Impl(columns, rows)
		}

		@JvmStatic
		fun unmodifiable(parent: Boolean2DArray): Boolean2DArray {
			if (parent is View)
				return parent
			else
				return View(parent)
		}

		/**
		 * Copies specified 2D array.
		 *
		 * This method may be more efficient than doing it manually,
		 * and semantically is equivalent to next code:
		 *
		 * ```kotlin
		 * val result = allocate(value.columns, value.rows)
		 * result.load(value)
		 * return result
		 * ```
		 */
		@JvmStatic
		fun copy(value: Boolean2DArray): Boolean2DArray {
			if (value is View) {
				return copy(value.parent)
			} else if (value is Impl) {
				return Impl(value)
			} else {
				val result = Impl(value.columns, value.rows)
				result.load(value)
				return result
			}
		}

		@JvmField
		val ZERO: Boolean2DArray = Impl(0, 0)
	}
}
