package ru.dbotthepony.kommons.io

import ru.dbotthepony.kommons.util.IStruct2b
import ru.dbotthepony.kommons.util.IStruct2d
import ru.dbotthepony.kommons.util.IStruct2f
import ru.dbotthepony.kommons.util.IStruct2i
import ru.dbotthepony.kommons.util.IStruct2l
import ru.dbotthepony.kommons.util.IStruct2s
import ru.dbotthepony.kommons.util.IStruct3b
import ru.dbotthepony.kommons.util.IStruct3d
import ru.dbotthepony.kommons.util.IStruct3f
import ru.dbotthepony.kommons.util.IStruct3i
import ru.dbotthepony.kommons.util.IStruct3l
import ru.dbotthepony.kommons.util.IStruct3s
import ru.dbotthepony.kommons.util.IStruct4b
import ru.dbotthepony.kommons.util.IStruct4d
import ru.dbotthepony.kommons.util.IStruct4f
import ru.dbotthepony.kommons.util.IStruct4i
import ru.dbotthepony.kommons.util.IStruct4l
import ru.dbotthepony.kommons.util.IStruct4s
import ru.dbotthepony.kommons.util.KOptional
import java.io.DataOutput
import java.io.OutputStream
import java.io.RandomAccessFile
import java.math.BigDecimal
import java.util.*
import java.util.function.IntConsumer

fun OutputStream.writeBigDecimal(value: BigDecimal) {
	writeInt(value.scale())
	val bytes = value.unscaledValue().toByteArray()
	writeVarInt(bytes.size)
	write(bytes)
}

fun <S : OutputStream, V> S.writeCollection(collection: Collection<V>, writer: S.(V) -> Unit) {
	writeVarInt(collection.size)

	for (value in collection) {
		writer(this, value)
	}
}

fun OutputStream.writeInt(value: Int) {
	if (this is DataOutput) {
		writeInt(value)
		return
	}

	write(value ushr 24)
	write(value ushr 16)
	write(value ushr 8)
	write(value)
}

fun OutputStream.writeShort(value: Int) {
	if (this is DataOutput) {
		writeShort(value)
		return
	}

	write(value ushr 8)
	write(value)
}

fun OutputStream.writeLong(value: Long) {
	if (this is DataOutput) {
		writeLong(value)
		return
	}

	write((value ushr 48).toInt())
	write((value ushr 40).toInt())
	write((value ushr 32).toInt())
	write((value ushr 24).toInt())
	write((value ushr 16).toInt())
	write((value ushr 8).toInt())
	write(value.toInt())
}

fun OutputStream.writeFloat(value: Float) = writeInt(value.toBits())
fun OutputStream.writeDouble(value: Double) = writeLong(value.toBits())

private inline fun writeVarLong(write: (Int) -> Unit, value: Long): Int {
	var length = 9

	while (length > 0) {
		val octetsPresent = (value ushr (length * 7)) and 0x7FL

		if (octetsPresent != 0L) {
			break
		} else {
			length--
		}
	}

	for (octet in length downTo 1) {
		write(((value ushr (octet * 7)) and 0x7FL).toInt() or 0x80)
	}

	write((value and 0x7FL).toInt())

	return length + 1
}

private inline fun writeVarInt(write: (Int) -> Unit, value: Int): Int {
	return writeVarLong(write, value.toLong())
}

private fun Int.toSignedVar(): Int {
	if (this >= 0)
		return this shl 1
	else
		return ((-this - 1) shl 1) or 1
}

private fun Long.toSignedVar(): Long {
	if (this >= 0)
		return this shl 1
	else
		return ((-this - 1) shl 1) or 1L
}

/**
 * Writes unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.writeVarLong(value: Long) = writeVarLong(::write, value)

/**
 * Writes unsigned variable length integer, with Big-endian byte order
 */
fun RandomAccessFile.writeVarInt(value: Int) = writeVarInt(::write, value)

/**
 * Writes unsigned variable length integer, with Big-endian byte order
 */
fun OutputStream.writeVarLong(value: Long) = writeVarLong(::write, value)

/**
 * Writes unsigned variable length integer, with Big-endian byte order
 */
fun OutputStream.writeVarInt(value: Int) = writeVarInt(::write, value)


/**
 * Writes signed variable length integer, with Big-endian byte order. Sign is encoded as first bit.
 */
fun RandomAccessFile.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())

/**
 * Writes signed variable length integer, with Big-endian byte order. Sign is encoded as first bit.
 */
fun RandomAccessFile.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())

/**
 * Writes signed variable length integer, with Big-endian byte order. Sign is encoded as first bit.
 */
fun OutputStream.writeSignedVarLong(value: Long) = writeVarLong(::write, value.toSignedVar())

/**
 * Writes signed variable length integer, with Big-endian byte order. Sign is encoded as first bit.
 */
fun OutputStream.writeSignedVarInt(value: Int) = writeVarInt(::write, value.toSignedVar())

fun OutputStream.writeBinaryString(input: String) {
	val bytes = input.encodeToByteArray()
	writeVarInt(bytes.size)
	write(bytes)
}

fun OutputStream.writeUTF(value: String) {
	write(value.toByteArray().also { check(!it.any { it.toInt() == 0 }) { "Provided UTF string contains NUL" } })
	write(0)
}

fun OutputStream.writeUUID(value: UUID) {
	writeLong(value.mostSignificantBits)
	writeLong(value.leastSignificantBits)
}

fun OutputStream.writeStruct2i(value: IStruct2i) {
	writeInt(value.component1())
	writeInt(value.component2())
}

fun OutputStream.writeStruct2f(value: IStruct2f) {
	writeFloat(value.component1())
	writeFloat(value.component2())
}

fun OutputStream.writeStruct2d(value: IStruct2d) {
	writeDouble(value.component1())
	writeDouble(value.component2())
}

fun OutputStream.writeStruct2l(value: IStruct2l) {
	writeLong(value.component1())
	writeLong(value.component2())
}

fun OutputStream.writeStruct2s(value: IStruct2s) {
	writeShort(value.component1().toInt() and 0xFFFF)
	writeShort(value.component2().toInt() and 0xFFFF)
}

fun OutputStream.writeStruct2b(value: IStruct2b) {
	writeShort(value.component1().toInt() and 0xFF)
	writeShort(value.component2().toInt() and 0xFF)
}

fun OutputStream.writeStruct3i(value: IStruct3i) {
	writeInt(value.component1())
	writeInt(value.component2())
	writeInt(value.component3())
}

fun OutputStream.writeStruct3f(value: IStruct3f) {
	writeFloat(value.component1())
	writeFloat(value.component2())
	writeFloat(value.component3())
}

fun OutputStream.writeStruct3d(value: IStruct3d) {
	writeDouble(value.component1())
	writeDouble(value.component2())
	writeDouble(value.component3())
}

fun OutputStream.writeStruct3l(value: IStruct3l) {
	writeLong(value.component1())
	writeLong(value.component2())
	writeLong(value.component3())
}

fun OutputStream.writeStruct3s(value: IStruct3s) {
	writeShort(value.component1().toInt() and 0xFFFF)
	writeShort(value.component2().toInt() and 0xFFFF)
	writeShort(value.component3().toInt() and 0xFFFF)
}

fun OutputStream.writeStruct3b(value: IStruct3b) {
	writeShort(value.component1().toInt() and 0xFF)
	writeShort(value.component2().toInt() and 0xFF)
	writeShort(value.component3().toInt() and 0xFF)
}

fun OutputStream.writeStruct4i(value: IStruct4i) {
	writeInt(value.component1())
	writeInt(value.component2())
	writeInt(value.component3())
	writeInt(value.component4())
}

fun OutputStream.writeStruct4f(value: IStruct4f) {
	writeFloat(value.component1())
	writeFloat(value.component2())
	writeFloat(value.component3())
	writeFloat(value.component4())
}

fun OutputStream.writeStruct4d(value: IStruct4d) {
	writeDouble(value.component1())
	writeDouble(value.component2())
	writeDouble(value.component3())
	writeDouble(value.component4())
}

fun OutputStream.writeStruct4l(value: IStruct4l) {
	writeLong(value.component1())
	writeLong(value.component2())
	writeLong(value.component3())
	writeLong(value.component4())
}

fun OutputStream.writeStruct4s(value: IStruct4s) {
	writeShort(value.component1().toInt() and 0xFFFF)
	writeShort(value.component2().toInt() and 0xFFFF)
	writeShort(value.component3().toInt() and 0xFFFF)
	writeShort(value.component4().toInt() and 0xFFFF)
}

fun OutputStream.writeStruct4b(value: IStruct4b) {
	writeShort(value.component1().toInt() and 0xFF)
	writeShort(value.component2().toInt() and 0xFF)
	writeShort(value.component3().toInt() and 0xFF)
	writeShort(value.component4().toInt() and 0xFF)
}

fun <S : OutputStream, K, V> S.writeMap(map: Map<K, V>, keyWriter: S.(K) -> Unit, valueWriter: S.(V) -> Unit) {
	writeVarInt(map.size)

	for ((k, v) in map.entries) {
		keyWriter(this, k)
		valueWriter(this, v)
	}
}

fun <S : OutputStream, V> S.writeKOptional(value: KOptional<V>, writer: S.(V) -> Unit) {
	write(if (value.isPresent) 1 else 0)

	if (value.isPresent) {
		writer(this, value.value)
	}
}

fun <S : OutputStream, V : Any> S.writeOptional(value: Optional<V>, writer: S.(V) -> Unit) {
	write(if (value.isPresent) 1 else 0)

	if (value.isPresent) {
		writer(this, value.get())
	}
}

/**
 * Writes array's size as variable length integer and then writes its contents
 */
fun OutputStream.writeByteArray(value: ByteArray) {
	writeVarInt(value.size)
	write(value)
}

/**
 * Writes array's size as variable length integer and then writes its contents
 */
fun OutputStream.writeByteArray(value: ByteArray, offset: Int, length: Int) {
	Objects.checkFromIndexSize(offset, length, value.size)
	writeVarInt(length)
	write(value, offset, length)
}
