package ru.dbotthepony.kommons.util

import java.util.concurrent.CopyOnWriteArrayList
import java.util.function.Consumer

interface Listenable<V> {
	/**
	 * Listener token, allows to remove listener from subscriber list
	 */
	fun interface L {
		/**
		 * Removes this listener
		 */
		fun remove()
	}

	fun addListener(listener: Runnable): L {
		return addListener(Consumer { listener.run() })
	}

	fun addListener(listener: Consumer<V>): L

	/**
	 * Adding and removing listeners is thread-safe.
	 */
	class Impl<V> : Listenable<V>, Consumer<V> {
		private inner class L(val callback: Consumer<V>) : Listenable.L {
			init {
				subscribers.add(this)
			}

			override fun remove() {
				subscribers.remove(this)
			}
		}

		private val subscribers = CopyOnWriteArrayList<L>()

		fun clear() {
			subscribers.clear()
		}

		override fun addListener(listener: Consumer<V>): Listenable.L {
			return L(listener)
		}

		override fun accept(t: V) {
			subscribers.forEach { it.callback.accept(t) }
		}
	}

	class Void : Listenable<Unit>, Runnable {
		private inner class L(val callback: Runnable) : Listenable.L {
			private var isRemoved = false

			init {
				subscribers.add(this)
			}

			override fun remove() {
				if (!isRemoved) {
					isRemoved = true
					subscribers.remove(this)
				}
			}
		}

		private val subscribers = CopyOnWriteArrayList<L>()

		fun clear() {
			subscribers.clear()
		}

		override fun addListener(listener: Runnable): Listenable.L {
			return L(listener)
		}

		override fun addListener(listener: Consumer<Unit>): Listenable.L {
			return L(Runnable { listener.accept(Unit) })
		}

		override fun run() {
			subscribers.forEach { it.callback.run() }
		}
	}

	companion object : Listenable<Nothing>, L {
		@Suppress("unchecked_cast")
		fun <T> empty(): Listenable<T> {
			return this as Listenable<T>
		}

		override fun remove() {}

		override fun addListener(listener: Consumer<Nothing>): L {
			return this
		}

		override fun addListener(listener: Runnable): L {
			return this
		}
	}
}
